#include "multiview_stereo.hpp"
#include "vr_app.hpp"

using namespace lava;

multiview_stereo::multiview_stereo(vr_app* app)
: stereo_strategy(app) {
}

multiview_stereo::~multiview_stereo() {
    pipeline_layout_->destroy();
}

bool multiview_stereo::create() {
    pipeline_layout_ = make_pipeline_layout();
    pipeline_layout_->add(app()->per_frame_descriptor());
    pipeline_layout_->add(app()->scene_descriptor());
    if (!pipeline_layout_->create(app()->device())) {
        return false;
    }

    create_render_pass(app()->right_eye_framebuffer(), app()->right_eye_framebuffer());

    return true;
}

void multiview_stereo::render(VkCommandBuffer command_buffer, lava::index frame) {
    // begin
    auto origin = area.get_origin();
    auto size = area.get_size();

    VkRenderPassBeginInfo const info{
        .sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO,
        .renderPass = vk_render_pass,
        .framebuffer = framebuffers[frame],
        .renderArea = { { origin.x, origin.y }, { size.x, size.y } },
        .clearValueCount = to_ui32(clear_values.size()),
        .pClearValues = clear_values.data(),
    };

    device().call().vkCmdBeginRenderPass(command_buffer, &info, VK_SUBPASS_CONTENTS_INLINE);

    // process
    ui32 count = 0;

    for (auto& subpass : subpasses) {
        if (count > 0)
            device().call().vkCmdNextSubpass(command_buffer, VK_SUBPASS_CONTENTS_INLINE);

        if (!subpass->activated())
            continue;

        subpass->process(command_buffer, area.get_size());

        ++count;
    }

    // end
    device().call().vkCmdEndRenderPass(command_buffer);
}

bool multiview_stereo::create_render_pass(const stereo_framebuffer::ptr& left_eye_framebuffer, const stereo_framebuffer::ptr& right_eye_framebuffer) {
    for (auto& attachment : attachments) {
        attachment_descriptions.push_back(attachment->get_description());
    }

    std::vector<VkSubpassDescription> subpass_descriptions;

    for (auto& subpass : subpasses)
        subpass_descriptions.push_back(subpass->get_description());

    std::vector<VkSubpassDependency> subpass_dependencies;

    for (auto& dependency : dependencies)
        subpass_dependencies.push_back(dependency->get_dependency());

    VkRenderPassCreateInfo const create_info{
        .sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO,
        .attachmentCount = to_ui32(attachment_descriptions.size()),
        .pAttachments = attachment_descriptions.data(),
        .subpassCount = to_ui32(subpass_descriptions.size()),
        .pSubpasses = subpass_descriptions.data(),
        .dependencyCount = to_ui32(subpass_dependencies.size()),
        .pDependencies = subpass_dependencies.data(),
    };

    if (!check(device->call().vkCreateRenderPass(device->get(), &create_info, memory::alloc(), &vk_render_pass))) {
        log()->error("create render pass");
        return false;
    }

    return true;
}
