#include "naive_stereo.hpp"
#include "vr_app.hpp"

using namespace lava;

naive_stereo::naive_stereo(vr_app* app) : stereo_strategy(app) {
}

naive_stereo::~naive_stereo() {
    pipeline_layout_->destroy();
    render_passes_[0].render_pass->destroy();
    render_passes_[0].pipeline->destroy();
    render_passes_[1].render_pass->destroy();
    render_passes_[1].pipeline->destroy();
}

bool naive_stereo::create() {
    pipeline_layout_ = make_pipeline_layout();
    pipeline_layout_->add(app()->per_frame_descriptor());
    pipeline_layout_->add(app()->scene()->material_descriptor);
    pipeline_layout_->add(app()->scene()->mesh_descriptor);
    if (!pipeline_layout_ ->create(app()->device())) {
        return false;
    }

    render_passes_[0] = create_render_pass(app()->left_eye_framebuffer());
    render_passes_[1] = create_render_pass(app()->right_eye_framebuffer());

    return true;
}

void naive_stereo::render(VkCommandBuffer command_buffer, lava::index frame) {
    for (auto& eye_pass : render_passes_) {
        eye_pass.render_pass->process(command_buffer, 0);
    }

    frame_capture().capture_image(command_buffer, app()->left_eye_framebuffer()->color_image, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, "left_eye");
    frame_capture().capture_image(command_buffer, app()->right_eye_framebuffer()->color_image, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL, "right_eye");

    pass_timer().begin_pass(command_buffer, "left_eye");
    render_passes_[0].render_pass->process(command_buffer, 0);
    pass_timer().end_pass(command_buffer);

    pass_timer().begin_pass(command_buffer, "right_eye");
    render_passes_[1].render_pass->process(command_buffer, 0);
    pass_timer().end_pass(command_buffer);
}

naive_stereo::eye_render_pass naive_stereo::create_render_pass(const stereo_framebuffer::ptr& framebuffer) {
    eye_render_pass render_pass;
    render_pass.pipeline = make_graphics_pipeline(app()->device());

    scene::set_vertex_input(render_pass.pipeline.get());

    const VkPipelineColorBlendAttachmentState blend_state = {
        .colorWriteMask = VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT
    };
    render_pass.pipeline->add_color_blend_attachment(blend_state);
    render_pass.pipeline->set_depth_test_and_write(true, true);
    render_pass.pipeline->set_depth_compare_op(VK_COMPARE_OP_LESS);


    render_pass.pipeline->on_process = [this, framebuffer](VkCommandBuffer command_buffer) {
        pipeline_layout_->bind(command_buffer, framebuffer->per_frame_descriptor_set, 0);

        for (size_t i = 0; i < app()->scene()->meshes.size(); ++i) {
            const auto& mesh = app()->scene()->meshes[i];
            const auto& material = app()->scene()->materials[app()->scene()->meshes[i].material_index];
            pipeline_layout_->bind(command_buffer, material->descriptor_set, 1);
            pipeline_layout_->bind(command_buffer, mesh.descriptor_set, 2);

            mesh.mesh->bind_draw(command_buffer);
        }
    };
    render_pass.pipeline->set_layout(pipeline_layout_);

    if (!render_pass.pipeline->add_shader(file_data("dpr/naive_stereo_vertex.spirv"), VK_SHADER_STAGE_VERTEX_BIT)) {
        log()->error("Failed to load shader");
        throw std::runtime_error("Failed to load shader");
    }

    if (!render_pass.pipeline->add_shader(file_data("dpr/naive_stereo_fragment.spirv"), VK_SHADER_STAGE_FRAGMENT_BIT)) {
        log()->error("Failed to load shader");
        throw std::runtime_error("Failed to load shader");
    }
    render_pass.render_pass = make_render_pass(app()->device());
    render_pass.render_pass->set_clear_color();

    subpass::ptr subpass = make_subpass();

    VkAttachmentReference color_attachment;
    color_attachment.attachment = 0;
    color_attachment.layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
    subpass->set_color_attachment(color_attachment);

    VkAttachmentReference depth_stencil_attachment;
    depth_stencil_attachment.attachment = 1;
    depth_stencil_attachment.layout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
    subpass->set_depth_stencil_attachment(depth_stencil_attachment);

    render_pass.render_pass->add(framebuffer->color_attachment);
    render_pass.render_pass->add(framebuffer->depth_stencil_attachment);
    render_pass.render_pass->add(subpass);

    subpass_dependency::ptr dependency = make_subpass_dependency(VK_SUBPASS_EXTERNAL, 0);
    // wait for previous fragment shader to finish reading before clearing attachments
    dependency->set_stage_mask(
        VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT,
        VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT | VK_PIPELINE_STAGE_EARLY_FRAGMENT_TESTS_BIT);
    // we need a memory barrier because this isn't a standard write-after-read hazard
    // subpass deps have an implicit attachment layout transition, so the dst access mask must be correct
    dependency->set_access_mask(0,
                                VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_READ_BIT | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT);
    render_pass.render_pass->add(dependency);

    dependency = make_subpass_dependency(render_pass.render_pass->get_subpass_count() - 1, VK_SUBPASS_EXTERNAL);
    // don't run any fragment shader (sample attachments) before we're done writing to attachments
    dependency->set_stage_mask(VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT | VK_PIPELINE_STAGE_LATE_FRAGMENT_TESTS_BIT,
                               VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT);
    // make attachment writes visible to subsequent reads
    dependency->set_access_mask(VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT,
                                VK_ACCESS_SHADER_READ_BIT);
    render_pass.render_pass->add(dependency);

    rect area = { glm::vec2(0.0f, 0.0f), framebuffer->color_image->get_size() };
    VkImageViews views {
        framebuffer->color_image->get_view(),
        framebuffer->depth_stencil_image->get_view(),
    };
    render_pass.render_pass->create({ views }, area);

    render_pass.pipeline->create(render_pass.render_pass->get());
    render_pass.render_pass->add_front(render_pass.pipeline);

    return render_pass;
}
