#include "pass_timer.hpp"
#include <fstream>
#include <algorithm>
#include <chrono>
#include <filesystem>

pass_timer::pass_timer(const std::string& base_directory) : base_directory(base_directory)
{

}

bool pass_timer::create(lava::device_ptr device, uint32_t frame_count, uint32_t pass_count)
{
    if(!this->load_timestamp_specs(device))
    {
        return false;
    }

    VkQueryPoolCreateInfo pool_create_info;
    pool_create_info.sType = VK_STRUCTURE_TYPE_QUERY_POOL_CREATE_INFO;
    pool_create_info.pNext = nullptr;
    pool_create_info.flags = 0;
    pool_create_info.queryType = VK_QUERY_TYPE_TIMESTAMP;
    pool_create_info.queryCount = pass_count * 2;
    pool_create_info.pipelineStatistics = 0;

    for(uint32_t frame = 0; frame < frame_count; frame++)
    {
        VkQueryPool query_pool = nullptr;
        if(vkCreateQueryPool(device->get(), &pool_create_info, lava::memory::alloc(), &query_pool) != VK_SUCCESS)
        {
            lava::log()->error("can't create query pool for pass timer!");

            this->destroy(device);

            return false;
        }

        this->query_pools.push_back(query_pool);
    }

    this->frame_count = frame_count;
    this->timestamp_count = pass_count * 2;

    this->current_frame = 0;
    this->current_timestamp = 0;

    this->timestamp_zero = {};

    return true;
}

void pass_timer::destroy(lava::device_ptr device)
{
    for(const VkQueryPool& pool : this->query_pools)
    {
        vkDestroyQueryPool(device->get(), pool, lava::memory::alloc());
    }

    this->query_pools.clear();

    this->frame_count = 0;
    this->timestamp_count = 0;

    this->current_frame = 0;
    this->current_timestamp = 0;

    this->timestamp_zero = {};
}

bool pass_timer::write_to_file(lava::device_ptr device)
{
    if(!this->wait_timestamps(device))
    {
        return false;
    }

    if(!std::filesystem::exists(this->base_directory))
    {
        std::filesystem::create_directory(this->base_directory);
    }

    return this->write_timestamps();
}

void pass_timer::next_frame(VkCommandBuffer command_buffer, lava::index frame, lava::device_ptr device)
{
    this->evaluate_wating_passes(device, frame, false);

    this->current_frame = frame;
    this->current_timestamp = 0;

    //Rest all timestamps to the unavailable state for the upcomming frame
    VkQueryPool current_pool = this->query_pools[frame];
    vkCmdResetQueryPool(command_buffer, current_pool, 0, this->timestamp_count);
}

void pass_timer::begin_pass(VkCommandBuffer command_buffer, const std::string& pass_name)
{
    if(this->current_timestamp >= this->timestamp_count)
    {
        lava::log()->error("not enough timestamps available. please increase the number of passes for the pass timer");

        return;
    }

    pass_wait_info wait_info;
    wait_info.name = pass_name;
    wait_info.frame_index = this->current_frame;
    wait_info.timestamp_index = this->current_timestamp;

    this->wait_infos.push_back(wait_info);

    VkQueryPool current_pool = this->query_pools[this->current_frame];
    vkCmdWriteTimestamp(command_buffer, VK_PIPELINE_STAGE_ALL_COMMANDS_BIT, current_pool, this->current_timestamp);

    this->current_timestamp++;
}

void pass_timer::end_pass(VkCommandBuffer command_buffer)
{
    if(this->current_timestamp >= this->timestamp_count)
    {
        lava::log()->error("not enough timestamps available. please increase the number of passes for the pass timer");

        return;
    }

    VkQueryPool current_pool = this->query_pools[this->current_frame];
    vkCmdWriteTimestamp(command_buffer, VK_PIPELINE_STAGE_ALL_COMMANDS_BIT, current_pool, this->current_timestamp);

    this->current_timestamp++;
}

bool pass_timer::load_timestamp_specs(lava::device_ptr device)
{
    uint32_t property_count = 0;
    vkGetPhysicalDeviceQueueFamilyProperties(device->get_vk_physical_device(), &property_count, nullptr);

    std::vector<VkQueueFamilyProperties> property_list(property_count);
    vkGetPhysicalDeviceQueueFamilyProperties(device->get_vk_physical_device(), &property_count, property_list.data());

    const VkPhysicalDeviceProperties& device_properties = device->get_properties();
    lava::queue::ref graphics_queue = device->get_graphics_queue();
    const VkQueueFamilyProperties& graphics_queue_properties = property_list[graphics_queue.family];
    
    if(graphics_queue_properties.timestampValidBits == 0)
    {
        lava::log()->error("the graphics queue does not support timestamps");

        return false;
    }

    this->timestamp_bits = graphics_queue_properties.timestampValidBits;
    this->timestamp_period = device_properties.limits.timestampPeriod;

    return true;
}

bool pass_timer::evaluate_wating_passes(lava::device_ptr device, lava::index frame, bool wait)
{
    VkQueryPool current_pool = this->query_pools[frame];
    frame_statistic frame_statistic;

    for(const pass_wait_info& wait_info : this->wait_infos)
    {
        if(wait_info.frame_index == frame)
        {
            uint32_t flags = VK_QUERY_RESULT_64_BIT;

            if(wait)
            {
                flags |= VK_QUERY_RESULT_WAIT_BIT;
            }

            std::array<uint64_t, 2> timestamps;
            if(vkGetQueryPoolResults(device->get(), current_pool, wait_info.timestamp_index, 2, sizeof(uint64_t) * timestamps.size(), timestamps.data(), 0, flags) != VK_SUCCESS)
            {
                lava::log()->error("can't get the timestamp results for the pass timer");

                return false;
            }

            if(!this->timestamp_zero.has_value())
            {
                this->timestamp_zero = timestamps[0];
            }

            //Subtract the end timestamp form the start timestamp and multiply it with timestamp_period.
            //The result of this computation gives the pass time in nano seconds
            //In order to get milli seconds divide by 10^6
            double start_time = (double)((timestamps[0] - this->timestamp_zero.value()) * this->timestamp_period) / 1000000.0;
            double end_time = (double)((timestamps[1] - this->timestamp_zero.value()) * this->timestamp_period) / 1000000.0;
            double delta_time = (double)((timestamps[1] - timestamps[0]) * this->timestamp_period) / 1000000.0;

            pass_statistic pass_statistic;
            pass_statistic.name = wait_info.name;
            pass_statistic.start_time = start_time;
            pass_statistic.end_time = end_time;
            pass_statistic.delta_time = delta_time;

            frame_statistic.push_back(pass_statistic);
        }
    }

    this->frame_statistics.push_back(frame_statistic);

    this->wait_infos.erase(std::remove_if(this->wait_infos.begin(), this->wait_infos.end(), [=](const pass_wait_info& wait_info)
    {
        return wait_info.frame_index == frame;
    }), this->wait_infos.end());

    return true;
}

bool pass_timer::wait_timestamps(lava::device_ptr device)
{
    for(uint32_t offset = 0; offset < this->frame_count; offset++)
    {
        uint32_t frame = (this->current_frame + offset + 1) % this->frame_count;

        if(!this->evaluate_wating_passes(device, frame, true))
        {
            return false;
        }
    }

    return true;
}

bool pass_timer::write_timestamps()
{
    std::string file_name = this->build_file_name();

    std::fstream file;
    file.open(this->base_directory + "/" + file_name, std::ios::out);

    if(!file.good())
    {
        lava::log()->error("can't write to file '" + file_name + "' for pass statistic");

        return false;
    }

    std::vector<std::string> pass_name_list;

    for(const std::vector<pass_statistic>& frame : this->frame_statistics)
    {
        for(const pass_statistic& pass : frame)
        {
            bool found = false;

            for(const std::string& name : pass_name_list)
            {
                if(name == pass.name)
                {
                    found = true;

                    break;
                }
            }

            if(!found)
            {
                pass_name_list.push_back(pass.name);
            }
        }
    }

    for(uint32_t index = 0; index < pass_name_list.size(); index++)
    {
        file << pass_name_list[index] << "_start_time;";
        file << pass_name_list[index] << "_end_time;";
        file << pass_name_list[index] << "_delta_time";

        if(index < pass_name_list.size() - 1)
        {
            file << ";";
        }
    }

    file << std::endl;

    for(const std::vector<pass_statistic>& frame : this->frame_statistics)
    {
        if(frame.empty())
        {
            continue;
        }

        for(uint32_t index = 0; index < pass_name_list.size(); index++)
        {
            const pass_statistic* pass = nullptr;

            for(const pass_statistic& frame_pass : frame)
            {
                if(frame_pass.name == pass_name_list[index])
                {
                    pass = &frame_pass;

                    break;
                }
            }

            if(pass != nullptr)
            {
                file << pass->start_time << ";";
                file << pass->end_time << ";";
                file << pass->delta_time;
            }

            else
            {
                file << ";;";
            }

            if(index < pass_name_list.size() - 1)
            {
                file << ";";
            }
        }

        file << std::endl;
    }

    file.close();

    return true;
}

std::string pass_timer::build_file_name()
{
    time_t system_time = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
    tm local_time = *localtime(&system_time);

    std::string name = "statistic";
    name += "_" + std::to_string(local_time.tm_mday);
    name += "-" + std::to_string(local_time.tm_mon + 1);
    name += "-" + std::to_string(local_time.tm_year + 1900);
    name += "_" + std::to_string(local_time.tm_hour);
    name += "-" + std::to_string(local_time.tm_min);
    name += "-" + std::to_string(local_time.tm_sec);
    name += ".csv";

    return name;
}