/*
  With the pass timer it is possible to measure the time that a pass needs to be completed on the gpu.
  The pass timer can be accessed form the stereo_strategy class with the function pass_timer().
  To measure the time of a pass call begin_pass(...) before and end_pass(...) after the pass.
  The resulting statistic is then stored in a csv-file.
  Example:

   pass_timer().begin_pass(command_buffer, "pass_name");
    //messured commands 
   pass_timer().end_pass(command_buffer, "pass_name");
*/

#pragma once
#include <liblava/lava.hpp>
#include <vector>
#include <cstdint>
#include <string>
#include <optional>

struct pass_wait_info
{
    std::string name;
    uint32_t frame_index = 0;
    uint32_t timestamp_index = 0;
};

struct pass_statistic
{
    std::string name;
    double start_time = 0.0;
    double end_time = 0.0;
    double delta_time = 0.0;
};

typedef std::vector<pass_statistic> frame_statistic;

class pass_timer
{
public:
    pass_timer(const std::string& base_directory);

    bool create(lava::device_ptr device, uint32_t frame_count, uint32_t pass_count);
    void destroy(lava::device_ptr device);

    bool write_to_file(lava::device_ptr device);

    void next_frame(VkCommandBuffer command_buffer, lava::index frame, lava::device_ptr device);
    void begin_pass(VkCommandBuffer command_buffer, const std::string& pass_name);
    void end_pass(VkCommandBuffer command_buffer);

private:
    bool load_timestamp_specs(lava::device_ptr device);
    bool evaluate_wating_passes(lava::device_ptr device, lava::index frame, bool wait);

    bool wait_timestamps(lava::device_ptr device);
    bool write_timestamps();

    std::string build_file_name();

private:
    std::string base_directory = "";

    uint32_t frame_count = 0;
    uint32_t timestamp_count = 0;

    uint32_t current_frame = 0;
    uint32_t current_timestamp = 0;

    uint32_t timestamp_bits = 0; //number of bits used for a timestamp in case of a graphics queue
    uint32_t timestamp_period = 0; //one increment of the timestamp in nanoseconds

    std::optional<uint64_t> timestamp_zero; //the timestamp that defines the zero time-point

    std::vector<VkQueryPool> query_pools;
    std::vector<pass_wait_info> wait_infos;
    std::vector<std::vector<pass_statistic>> frame_statistics;
};