#include "scene.hpp"
#include <cmath>
#include <assimp/Importer.hpp>
#include <assimp/postprocess.h>
#include <assimp/scene.h>
#include <filesystem>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtx/string_cast.hpp>
#include <glm/gtx/quaternion.hpp>

using namespace lava;

void scene::set_vertex_input(lava::graphics_pipeline* pipeline) {
    pipeline->set_vertex_input_binding({ 0, sizeof(vertex), VK_VERTEX_INPUT_RATE_VERTEX });
    pipeline->set_vertex_input_attributes({
        { 0, 0, VK_FORMAT_R32G32B32_SFLOAT, to_ui32(offsetof(vertex, position)) },
        { 1, 0, VK_FORMAT_R32G32B32_SFLOAT, to_ui32(offsetof(vertex, normal)) },
        { 2, 0, VK_FORMAT_R32G32B32_SFLOAT, to_ui32(offsetof(vertex, color)) }, // color contains actually the tangents
        { 3, 0, VK_FORMAT_R32G32_SFLOAT, to_ui32(offsetof(vertex, uv))},
    });
}

void scene::update(float delta_time) {
    if (!active_animation_index_.has_value()) {
        return;
    }

    const auto active_animation = get_active_animation();
    current_time += delta_time * active_animation->ticks_per_second;
    if (loop_) {
      current_time = std::fmod(current_time, active_animation->duration);
    } else {
      current_time = std::min(current_time, active_animation->duration);
    }

    for (const auto& channel : active_animation->channels) {
      glm::vec3 position(0.0f);
      glm::quat rotation(1.0f, 0.0f, 0.0f, 0.0f);
      glm::vec3 scaling(1.0f);

      std::pair<float, glm::vec3> last_position_frame = channel.position_frames.front();
      for (const auto& position_frame : channel.position_frames) {
          if (position_frame.first > current_time) {
            const float t = (current_time - last_position_frame.first) / (position_frame.first - last_position_frame.first);
            position = glm::mix(last_position_frame.second, position_frame.second, t);
            break;
          }
          last_position_frame = position_frame;
      }
      std::pair<float, glm::quat> last_rotation_frame = channel.rotation_frames.front();
      for (const auto& rotation_frame : channel.rotation_frames) {
          if (rotation_frame.first > current_time) {
            const float t = (current_time - last_rotation_frame.first) / (rotation_frame.first - last_rotation_frame.first);
            rotation = glm::mix(last_rotation_frame.second, rotation_frame.second, t);
            break;
          }
          last_rotation_frame = rotation_frame;
      }
      std::pair<float, glm::vec3> last_scaling_frame = channel.scaling_frames.front();
      for (const auto& scaling_frame : channel.scaling_frames) {
          if (scaling_frame.first > current_time) {
            const float t = (current_time - last_scaling_frame.first) / (scaling_frame.first - last_scaling_frame.first);
            scaling = glm::mix(last_scaling_frame.second, scaling_frame.second, t);
            break;
          }
          last_scaling_frame = scaling_frame;
      }

      nodes[channel.node_index].current_local_transform =
        glm::translate(glm::mat4(1.0f), position) *
        glm::toMat4(rotation) *
        glm::scale(glm::mat4(1.0f), scaling);
    }

    for (auto& node : nodes) {
        if (node.parent_index != std::numeric_limits<size_t>::max()) {
            assert(node.parent_index < (&node - nodes.data()));
            node.current_global_transform = nodes[node.parent_index].current_global_transform * node.current_local_transform;
        } else {
            node.current_global_transform = node.current_local_transform;
        }
    }
    for (size_t i = 0; i < meshes.size(); ++i) {
        mesh_data[i].localToWorldSpace = nodes[meshes[i].node_index].current_global_transform;
    }
    std::memcpy(mesh_data_buffer->get_mapped_data(), mesh_data.data(), sizeof(glsl::MeshData) * meshes.size());
}

void scene::set_animation(size_t animation_index) {
  current_time = 0.0f;
  active_animation_index_ = animation_index;
  reset_transforms();
}

void scene::clear_animation() {
    current_time = 0.0f;
    active_animation_index_.reset();
    reset_transforms();
}

void scene::reset_transforms() {
    for (auto& node : nodes) {
        node.current_local_transform = node.initial_local_transform;
        if (node.parent_index != std::numeric_limits<size_t>::max()) {
            assert(node.parent_index < (&node - nodes.data()));
            node.current_global_transform = nodes[node.parent_index].current_global_transform * node.current_local_transform;
        } else {
            node.current_global_transform = node.current_local_transform;
        }
    }
    for (size_t i = 0; i < meshes.size(); ++i) {
        mesh_data[i].localToWorldSpace = nodes[meshes[i].node_index].current_global_transform;
    }
    std::memcpy(mesh_data_buffer->get_mapped_data(), mesh_data.data(), sizeof(glsl::MeshData) * meshes.size());
}

namespace {

void insert_node(scene* scene, aiNode* assimp_node, size_t parent_index, std::map<std::string, std::size_t>* node_indices) {
  ::scene::node node {
      .current_global_transform = glm::mat4(1.0f),
      .parent_index = parent_index,
  };
  for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
          // Implicit transpose
          node.initial_local_transform[i][j] = assimp_node->mTransformation[j][i];
      }
  }
  const size_t node_index = scene->nodes.size();
  scene->nodes.push_back(node);
  node_indices->insert(std::make_pair(assimp_node->mName.C_Str(), node_index));
  for (unsigned int i = 0; i < assimp_node->mNumMeshes; ++i) {
    const size_t mesh_index = assimp_node->mMeshes[i];
    scene->meshes[mesh_index].node_index = node_index;
  }

  for (unsigned int i = 0; i < assimp_node->mNumChildren; ++i) {
    insert_node(scene, assimp_node->mChildren[i], node_index, node_indices);
  }
}

}

scene::ptr load_scene(lava::device_ptr device, lava::name filename) {
    Assimp::Importer importer;
    auto assimp_scene = importer.ReadFile(filename, 
        aiProcess_CalcTangentSpace       |
        aiProcess_ConvertToLeftHanded    |
        aiProcess_Triangulate            |
        aiProcess_JoinIdenticalVertices  |
        aiProcess_SortByPType);

    if (assimp_scene == nullptr) {
        log()->error("Failed to load {}", filename);
        return {};
    }


    auto scene = std::make_shared<::scene>();

    scene->material_descriptor = make_descriptor();
    scene->material_descriptor->add_binding(0, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT);
    scene->material_descriptor->add_binding(1, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT);
    // scene->material_descriptor->add_binding(2, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT);
    if (!scene->material_descriptor->create(device)) {
        return {};
    }
    scene->mesh_descriptor = make_descriptor();
    scene->mesh_descriptor ->add_binding(0, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_VERTEX_BIT);
    if (!scene->mesh_descriptor ->create(device)) {
        return {};
    }
    scene->dummy_texture = create_default_texture(device, {1, 1});

    scene->descriptor_pool_ = make_descriptor_pool();
    if (!scene->descriptor_pool_->create(device, {
            { VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 2 * assimp_scene->mNumMaterials },
            { VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 1 * assimp_scene->mNumMeshes}, 
        }, assimp_scene->mNumMaterials + assimp_scene->mNumMeshes)) {
        return {};
    }
    scene->meshes.resize(assimp_scene->mNumMeshes);

    std::map<std::string, size_t> node_indices;
    insert_node(scene.get(), assimp_scene->mRootNode, std::numeric_limits<size_t>::max(), &node_indices);

    auto assimp_vector3_to_glm = [](aiVector3D v) {
        return glm::vec3(v.x, v.y, v.z);
    };
    auto assimp_quaternion_to_glm = [](aiQuaternion v) {
        return glm::quat(v.w, v.x, v.y, v.z);
    };

    for (unsigned int i = 0; i < assimp_scene->mNumCameras; ++i) {
        const auto assimp_camera = assimp_scene->mCameras[i];

        if (!node_indices.contains(assimp_camera->mName.C_Str())) {
            log()->warn("Unknown node: {}", assimp_camera->mName.C_Str());
            continue;
        }

        scene_camera camera;
        camera.node_index = node_indices[assimp_camera->mName.C_Str()];
        camera.view_matrix = glm::lookAt(
            assimp_vector3_to_glm(assimp_camera->mPosition),
            assimp_vector3_to_glm(assimp_camera->mLookAt),
            assimp_vector3_to_glm(assimp_camera->mUp)
        );
        camera.name = assimp_camera->mName.C_Str();
        camera.near_plane = assimp_camera->mClipPlaneNear;
        camera.far_plane = assimp_camera->mClipPlaneFar;
        scene->cameras.push_back(std::move(camera));
    }

    for (unsigned int i = 0; i < assimp_scene->mNumAnimations; ++i) {
        const auto assimp_animation = assimp_scene->mAnimations[i];
        scene_animation animation;

        animation.duration = assimp_animation->mDuration;
        animation.ticks_per_second = assimp_animation->mTicksPerSecond;
        animation.name = assimp_scene->mName.C_Str();

        for (unsigned int j = 0; j < assimp_animation->mNumChannels; ++j) {
            const auto assimp_channel = assimp_animation->mChannels[j];

            scene_animation_channel channel;

            if (!node_indices.contains(assimp_channel->mNodeName.C_Str())) {
                log()->warn("Unknown node: {}", assimp_channel->mNodeName.C_Str());
                continue;
            }

            channel.node_index = node_indices[assimp_channel->mNodeName.C_Str()];

            channel.position_frames.reserve(assimp_channel->mNumPositionKeys);
            for (unsigned int k = 0; k < assimp_channel->mNumPositionKeys; ++k) {
                const auto key = assimp_channel->mPositionKeys[k];
                channel.position_frames.push_back(std::make_pair(key.mTime, assimp_vector3_to_glm(key.mValue)));
            }

            channel.rotation_frames.reserve(assimp_channel->mNumRotationKeys);
            for (unsigned int k = 0; k < assimp_channel->mNumRotationKeys; ++k) {
                const auto key = assimp_channel->mRotationKeys[k];
                channel.rotation_frames.push_back(std::make_pair(key.mTime, assimp_quaternion_to_glm(key.mValue)));
            }
            
            channel.scaling_frames.reserve(assimp_channel->mNumScalingKeys);
            for (unsigned int k = 0; k < assimp_channel->mNumScalingKeys; ++k) {
                const auto key = assimp_channel->mScalingKeys[k];
                channel.scaling_frames.push_back(std::make_pair(key.mTime, assimp_vector3_to_glm(key.mValue)));
            }

            animation.channels.push_back(std::move(channel));
        }

        scene->animations.push_back(std::move(animation));
    }

    scene->mesh_data.resize(assimp_scene->mNumMeshes, glsl::MeshData { .localToWorldSpace = glm::mat4(1.0f) });
    scene->mesh_data_buffer = make_buffer();
    if (!scene->mesh_data_buffer->create_mapped(device, scene->mesh_data.data(), scene->meshes.size() * sizeof(glsl::MeshData), VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)) {
        throw std::runtime_error("Failed to create mesh data buffer");
    }


    std::vector<VkWriteDescriptorSet> descriptor_set_writes;
    for (unsigned int i = 0; i < assimp_scene->mNumMeshes; ++i) {
        auto assimp_mesh = assimp_scene->mMeshes[i];

        mesh_data data;
        data.vertices.resize(assimp_mesh->mNumVertices);
        for (unsigned int vertex_index = 0; vertex_index < assimp_mesh->mNumVertices; ++vertex_index) {
            data.vertices[vertex_index].position = assimp_vector3_to_glm(assimp_mesh->mVertices[vertex_index]);
            if (assimp_mesh->HasNormals()) {
                data.vertices[vertex_index].normal = assimp_vector3_to_glm(assimp_mesh->mNormals[vertex_index]);
            }
            if (assimp_mesh->HasTangentsAndBitangents()) {
                data.vertices[vertex_index].color = glm::vec4(assimp_vector3_to_glm(assimp_mesh->mTangents[vertex_index]), 0.0f);
            }
            if (assimp_mesh->HasTextureCoords(0)) {
                data.vertices[vertex_index].uv = glm::vec2(assimp_vector3_to_glm(assimp_mesh->mTextureCoords[0][vertex_index]));
            }
        }
        data.indices.reserve(assimp_mesh->mNumFaces * 3);
        for (unsigned int face_index = 0; face_index < assimp_mesh->mNumFaces; ++face_index) {
            auto face = assimp_mesh->mFaces[face_index];
            if (face.mNumIndices != 3) {
                break;
            }
            data.indices.push_back(face.mIndices[0]);
            data.indices.push_back(face.mIndices[1]);
            data.indices.push_back(face.mIndices[2]);
        }
        if (data.indices.size() == 0) {
            continue;
        }

        scene->meshes[i].mesh = make_mesh();
        scene->meshes[i].mesh->add_data(data);
        scene->meshes[i].mesh->create(device);
        scene->meshes[i].material_index = assimp_mesh->mMaterialIndex;
        scene->meshes[i].descriptor_set = scene->mesh_descriptor->allocate_set(scene->descriptor_pool_->get());
        VkDescriptorBufferInfo buffer_info = {
            .buffer = scene->mesh_data_buffer->get(),
            .offset = sizeof(glsl::MeshData) * i,
            .range = sizeof(glsl::MeshData),
        };
        device->vkUpdateDescriptorSets({
            {
                .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
                .dstSet = scene->meshes[i].descriptor_set,
                .dstBinding = 0,
                .descriptorCount = 1,
                .descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                .pBufferInfo = &buffer_info,
            }
        });
    }

    const auto base_path = std::filesystem::path(filename).parent_path();
    for (unsigned int i = 0; i < assimp_scene->mNumMaterials; ++i) {
        auto assimp_material = assimp_scene->mMaterials[i];

        scene_material::ptr material = std::make_shared<scene_material>();
        material->descriptor_set = scene->material_descriptor->allocate_set(scene->descriptor_pool_->get());

        aiString diffuse_texture_path;
        assimp_material->GetTexture(aiTextureType_DIFFUSE, 0, &diffuse_texture_path);
        if (diffuse_texture_path.length > 0) {
            const auto texture_path = base_path / diffuse_texture_path.C_Str();
            std::string texture_path_string = texture_path.string();
            std::replace(texture_path_string.begin(), texture_path_string.end(), '\\', '/');
            material->diffuse = load_texture(device, texture_path_string, VK_FORMAT_UNDEFINED);

            descriptor_set_writes.push_back({
                .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
                .dstSet = material->descriptor_set,
                .dstBinding = 0,
                .descriptorCount = 1,
                .descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                .pImageInfo = material->diffuse->get_descriptor_info()
            });
        } else {
            descriptor_set_writes.push_back({
                .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
                .dstSet = material->descriptor_set,
                .dstBinding = 0,
                .descriptorCount = 1,
                .descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                .pImageInfo = scene->dummy_texture->get_descriptor_info()
            });
        }

        aiString normal_map_path;
        assimp_material->GetTexture(aiTextureType_NORMALS, 0, &normal_map_path);
        if (normal_map_path.length > 0) {
            auto texture_path = base_path / normal_map_path.C_Str();
            std::string texture_path_string = texture_path.string();
            std::replace(texture_path_string.begin(), texture_path_string.end(), '\\', '/');
            log()->info("Load normal map: {}", texture_path_string);
            material->normal = load_texture(device, texture_path_string, VK_FORMAT_UNDEFINED);
            if (material->normal->get_size() == glm::uvec2(1, 1)) {
              material->normal.reset();
              descriptor_set_writes.push_back({
                  .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
                  .dstSet = material->descriptor_set,
                  .dstBinding = 1,
                  .descriptorCount = 1,
                  .descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                  .pImageInfo = scene->dummy_texture->get_descriptor_info()
              });
            } else {
              descriptor_set_writes.push_back({
                  .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
                  .dstSet = material->descriptor_set,
                  .dstBinding = 1,
                  .descriptorCount = 1,
                  .descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                  .pImageInfo = material->normal->get_descriptor_info()
              });
            }
        } else {
            descriptor_set_writes.push_back({
                .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
                .dstSet = material->descriptor_set,
                .dstBinding = 1,
                .descriptorCount = 1,
                .descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
                .pImageInfo = scene->dummy_texture->get_descriptor_info()
            });
        }

        scene->materials.push_back(material);
    }
    device->vkUpdateDescriptorSets(descriptor_set_writes.size(), descriptor_set_writes.data());

    return scene;
}
