#include "vr_app.hpp"

#include <imgui.h>
#include <glm/gtx/string_cast.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace lava;

namespace {
std::vector<std::string> split_string(std::string_view string, char delimiter = ' ') {
    std::vector<std::string> strings;
    while (true) {
        auto next_occurence = string.find(delimiter);
        strings.emplace_back(string.substr(0, next_occurence));
        if (next_occurence == std::string_view::npos) {
            break;
        }
        string = string.substr(next_occurence + 1);
    }
    return strings;
}
}

vr_app::vr_app(name name, argh::parser cmd_line) {
    vr::HmdError hmd_error;
    vr_system.reset(VR_Init(&hmd_error, vr::EVRApplicationType::VRApplication_Scene));
    if (!vr_system) {
        if (log()) {
            log()->error("Failed to initialize OpenVR: {}", hmd_error);
        }
    }

    const auto pos_args = cmd_line.pos_args();
    if (pos_args.size() > 1) {
        scene_path_ = pos_args[1];
    }

    // Init app
    frame_config config(name, std::forward<argh::parser>(cmd_line));
    if (vr_system) {
        std::string required_instance_extensions_string;
        const auto required_buffer_size = vr::VRCompositor()->GetVulkanInstanceExtensionsRequired(nullptr, 0);
        required_instance_extensions_string.resize(required_buffer_size - 1); // required_buffers_size includes the null terminator
        vr::VRCompositor()->GetVulkanInstanceExtensionsRequired(required_instance_extensions_string.data(), required_instance_extensions_string.size());
        auto required_instance_extensions = split_string(required_instance_extensions_string);

        config.param.extensions.reserve(config.param.extensions.size() + required_instance_extensions.size());
        for (const auto& extension : required_instance_extensions) {
            config.param.extensions.push_back(extension.c_str());
        }
    }
    app_.emplace(config);

    // Create device
    {
        // VkPhysicalDevice physical_device;
        // static_assert(sizeof(VkPhysicalDevice) == sizeof(uint64_t));
        // vr_system->GetOutputDevice(reinterpret_cast<uint64_t*>(&physical_device), vr::ETextureType::TextureType_Vulkan, instance::get());

        std::vector<std::string> required_device_extensions;

        app_->manager.on_create_param = [&, this](device::create_param& device_param) {
            if (vr_system) {
                auto device = device_param.physical_device->get();

                std::string required_device_extensions_string;
                const auto required_buffer_size = vr::VRCompositor()->GetVulkanDeviceExtensionsRequired(device, nullptr, 0);
                required_device_extensions_string.resize(required_buffer_size - 1); // required_buffers_size includes the null terminator
                vr::VRCompositor()->GetVulkanDeviceExtensionsRequired(device, required_device_extensions_string.data(), required_device_extensions_string.size());

                required_device_extensions = split_string(required_device_extensions_string);
                device_param.extensions.reserve(device_param.extensions.size() + required_device_extensions.size());
                for (const auto& extension : required_device_extensions) {
                    device_param.extensions.push_back(extension.c_str());
                }
            }
            device_param.extensions.push_back("VK_NV_geometry_shader_passthrough");
            device_param.extensions.push_back("VK_NV_viewport_array2");
            device_param.features.fillModeNonSolid = true;
            device_param.features.multiViewport = true;
            device_param.features.geometryShader = true;
            device_param.features.tessellationShader = true;
        };

        // TODO: select proper device!
        app_->device = app_->create_device();
        if (!app_->device) {
            throw std::runtime_error("Failed to create vulkan device");
        }
    }
    app_->camera.movement_speed = 10.0f;

    app_->on_process = [this](VkCommandBuffer command_buffer, lava::index frame) {
        frame_capture_.next_frame(frame);
        pass_timer_.next_frame(command_buffer, frame, app_->device);

        VkImageSubresourceRange const image_range{
            .aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
            .levelCount = 1,
            .layerCount = 1,
        };
        insert_image_memory_barrier(device(), command_buffer, framebuffers_[0]->color_image->get(),
                            VK_ACCESS_TRANSFER_WRITE_BIT, VK_ACCESS_MEMORY_READ_BIT,
                            VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
                            VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
                            image_range);
        insert_image_memory_barrier(device(), command_buffer, framebuffers_[1]->color_image->get(),
                            VK_ACCESS_TRANSFER_WRITE_BIT, VK_ACCESS_MEMORY_READ_BIT,
                            VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
                            VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
                            image_range);
        VkImageSubresourceRange const depth_image_range{
            .aspectMask = VK_IMAGE_ASPECT_DEPTH_BIT,
            .levelCount = 1,
            .layerCount = 1,
        };
        insert_image_memory_barrier(device(), command_buffer, framebuffers_[0]->depth_stencil_image->get(),
                            VK_ACCESS_TRANSFER_WRITE_BIT, VK_ACCESS_MEMORY_READ_BIT,
                            VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
                            VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
                            depth_image_range);
        insert_image_memory_barrier(device(), command_buffer, framebuffers_[1]->depth_stencil_image->get(),
                            VK_ACCESS_TRANSFER_WRITE_BIT, VK_ACCESS_MEMORY_READ_BIT,
                            VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
                            VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
                            depth_image_range);
        stereo_strategies_[selected_strategy_]->render(command_buffer, frame);
        stereo_strategies_[selected_strategy_]->render_companion_window(command_buffer);
    };

    static const glm::dmat4 Y_FLIP(
        1.0, 0.0, 0.0, 0.0,
        0.0, -1.0, 0.0, 0.0,
        0.0, 0.0, 1.0, 0.0,
        0.0, 0.0, 0.0, 1.0);

    app_->on_update = [&](delta dt) {
        app_->camera.update_view(dt, app_->input.get_mouse_position());
        scene()->update(dt * animation_playback_speed_);

        glm::dmat4 view_matrix;
        if (selected_camera_.has_value()) {
          const auto& camera = scene()->cameras[*selected_camera_];
          view_matrix = Y_FLIP * glm::transpose(glm::dmat4(camera.view_matrix)) * glm::inverse(glm::dmat4(scene()->nodes[camera.node_index].current_global_transform));
        } else {
          view_matrix = glm::dmat4(camera().view) * Y_FLIP;
        }

        for (auto& framebuffer : framebuffers_) {
            const glm::dmat4 head_to_eye = framebuffer->per_frame_data.headToEye;
            const glm::dmat4 eye_view_matrix = head_to_eye * view_matrix;
            framebuffer->per_frame_data.view = eye_view_matrix;
            framebuffer->per_frame_data.viewProjection = glm::dmat4(framebuffer->per_frame_data.projection) * eye_view_matrix;
            std::memcpy(framebuffer->per_frame_ubo->get_mapped_data(), &framebuffer->per_frame_data, sizeof(framebuffer->per_frame_data));
        }
        stereo_strategies_[selected_strategy_]->update();
        frame_times_[frame_times_begin_] = dt * 1000;
        frame_times_begin_ = (frame_times_begin_ + 1) % frame_times_.size();
        return true;
    };

    app_->add_run_end([this]() {
        return true;
    });
}

bool vr_app::setup() {
    if (!app_->setup()) {
        return false;
    }

    frame_capture_.create(app_->target->get_frame_count());
    pass_timer_.create(app_->device, app_->target->get_frame_count(), 10);

    descriptor_pool_ = make_descriptor_pool();
    if (!descriptor_pool_->create(app_->device, {
            { VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 2 }, // Stereo images
            { VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 2 }, // Per frame data
        }, 3)) {
        return false;
    }

    per_frame_descriptor_ = make_descriptor();
    per_frame_descriptor_->add_binding(0, VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, VK_SHADER_STAGE_VERTEX_BIT | VK_SHADER_STAGE_FRAGMENT_BIT);
    if (!per_frame_descriptor_->create(device())) {
        return false;
    }

    glm::uvec2 render_target_size = { 1024, 1024 };
    if (vr_system) {
	    vr_system->GetRecommendedRenderTargetSize(&render_target_size.x, &render_target_size.y);
    }
    framebuffers_[0] = create_framebuffer(render_target_size, vr::EVREye::Eye_Left);
    framebuffers_[1] = create_framebuffer(render_target_size, vr::EVREye::Eye_Right);

    test_texture_ = create_default_texture(app_->device, { 1, 1 });
    // test_texture_ = load_texture(app_->device, "light/normal.png");
    if (!test_texture_) {
        log()->error("Failed to load test texture");
        return false;
    }
    app_->staging.add(test_texture_);

    quad_mesh_ = create_mesh(app_->device, mesh_type::quad);
    if (!quad_mesh_) {
        log()->error("Failed to create mesh");
        return false;
    }

    log()->info("Loading scene {}", scene_path_);
    scene_ = load_scene(app_->device, scene_path_.c_str());
    if (!scene_) {
        log()->error("Failed to load scene: {}!", scene_path_);
    } else {
        for (auto& material : scene_->materials) {
            if (material->diffuse) {
                staging().add(material->diffuse);
            }
            if (material->normal) {
                staging().add(material->normal);
            }
        }
        staging().add(scene_->dummy_texture);
        if (scene()->cameras.size() > 0) {
          selected_camera_ = 0;
          app_->camera.z_near = scene_->cameras[0].near_plane;
          app_->camera.z_far = scene_->cameras[0].far_plane;
        }
        if (scene()->animations.size() > 0) {
          scene()->set_animation(0);
        }
    }
    calculate_projection_matrices();

    app_->add_run_end([this]() {
        frame_capture_.destroy();
        pass_timer_.write_to_file(app_->device);
        pass_timer_.destroy(app_->device);

        framebuffers_[0]->destroy();
        framebuffers_[1]->destroy();
        test_texture_->destroy();
        quad_mesh_->destroy();
    });

    if (!setup_companion_window_pipeline()) {
        return false;
    }

    stereo_strategies_[0] = std::make_unique<naive_stereo>(this);
    if (!stereo_strategies_[0]->create()) {
        return false;
    }
    // multiview_stereo_ = std::make_unique<multiview_stereo>(this);
    // if (!multiview_stereo_->create()) {
    //     return false;
    // }
    stereo_strategies_[1] = std::make_unique<depth_peeling_reprojection_stereo>(this);
    if (!stereo_strategies_[1]->create()) {
        return false;
    }
    selected_strategy_ = 1;

    app_->imgui.on_draw = [&]() {
        ImGui::SetNextWindowPos(ImVec2(30, 30), ImGuiCond_FirstUseEver);
        ImGui::SetNextWindowSize(ImVec2(193, 90), ImGuiCond_FirstUseEver);

        ImGui::Begin("Application");
        if (ImGui::CollapsingHeader("Animation", ImGuiTreeNodeFlags_DefaultOpen)) {
          // Animation selector
          { 
            std::vector<const char*> items;
            items.reserve(scene()->animations.size() + 1);
            items.push_back("None");
            for (const auto& animation : scene()->animations) {
              items.push_back(animation.name.c_str());
            }
            const auto active_animation_index = scene()->get_active_animation_index();
            int selected_option = active_animation_index.has_value() ? *active_animation_index + 1 : 0;
            if (ImGui::Combo("Animation##active_animation", &selected_option, items.data(), items.size())) {
              if (selected_option == 0) {
                scene()->clear_animation();
              } else {
                scene()->set_animation(selected_option - 1);
              }
            }
          }
          auto active_animation = scene()->get_active_animation();
          if (active_animation != nullptr) {
              ImGui::SliderFloat("Time", &scene()->current_time, 0.0f, active_animation->duration);
              ImGui::DragFloat("Playback Speed", &animation_playback_speed_, 1.0f, -100.0f, 100.0);
              ImGui::Checkbox("Loop", &scene()->loop_);
          }
        }
        if (ImGui::CollapsingHeader("Camera", ImGuiTreeNodeFlags_DefaultOpen)) {
          std::vector<const char*> items;
          items.reserve(scene()->cameras.size() + 1);
          items.push_back("None");
          for (const auto& camera : scene()->cameras) {
            items.push_back(camera.name.c_str());
          }
          int selected_camra = selected_camera_.has_value() ? *selected_camera_ + 1 : 0;
          if (ImGui::Combo("Camera##selected_camera", &selected_camra, items.data(), items.size())) {
            if (selected_camra == 0) {
              selected_camera_.reset();
            } else {
              selected_camera_.emplace(selected_camra - 1);
              auto& camera = scene()->cameras[selected_camra - 1];
              app_->camera.z_near = camera.near_plane;
              app_->camera.z_far = camera.far_plane;
              calculate_projection_matrices();
            }
          }
          if (ImGui::DragFloat("Near Plane", &camera().z_near)) {
            calculate_projection_matrices();
          }
          if (ImGui::DragFloat("Far Plane", &camera().z_far)) {
            calculate_projection_matrices();
          }

          if (!selected_camera_.has_value()) {
              ImGui::DragFloat("Movement Speed", &camera().movement_speed, 1.0f, 0.0f, 100.0);
          }
        }

        if (ImGui::CollapsingHeader("Stereo", ImGuiTreeNodeFlags_DefaultOpen)) {
          std::array<const char*, 2> strategy_names = { "Naive Stereo", "Depth Peeling Reprojection" };
          ImGui::Combo("Stereo Strategy", &selected_strategy_, strategy_names.data(), strategy_names.size());

          ImGui::TreePush();
          stereo_strategies_[selected_strategy_]->draw_imgui();
          ImGui::TreePop();

        }

        if(ImGui::CollapsingHeader("Capture", ImGuiTreeNodeFlags_DefaultOpen))
        {
            bool active = frame_capture_.is_enabled();
            ImGui::Checkbox("Active", &active);

            if(active)
            {
                frame_capture_.enable();
            }

            else
            {
                frame_capture_.disable();
            }
        }

        if (ImGui::CollapsingHeader("Timings", ImGuiTreeNodeFlags_DefaultOpen)) {
          // if (ImGui::InputInt("Frame Time Smoothing", &frame_time_smoothing_)) {
          //   frame_time_smoothing_ = std::abs(frame_time_smoothing_);
          // }

          // std::array<float, FRAME_TIMES_BUFFER_SIZE> smooth_frame_times;
          // const int frame_time_smoothing_offset = frame_time_smoothing_ / 2;
          // for (int i = 0; i < FRAME_TIMES_BUFFER_SIZE; ++i) {
          //   float total = 0.0f;
          //   for (int j = 0; j < frame_time_smoothing_; ++j) {
          //     const int index = i + j - frame_time_smoothing_offset;
          //     total += frame_times_[std::clamp(index, 0, FRAME_TIMES_BUFFER_SIZE - 1)];
          //   }
          //   smooth_frame_times[i] = total / frame_time_smoothing_;
          // }

          // float min = std::numeric_limits<float>::infinity();
          // float max = -std::numeric_limits<float>::infinity();
          // for (const float time : frame_times_) {
          //     if (time < min) {
          //       min = time;
          //     }
          //     if (time > max) {
          //       max = time;
          //     }
          // }
          // std::string title = fmt::format("Frame Time (min: {}, max: {})###Frame Time", min, max);
          ImGui::PlotLines("Frame Times", frame_times_.data(), frame_times_.size(), frame_times_begin_, nullptr, 0.0, 11.11);
        }

        app_->draw_about(true);

        ImGui::End();
    };

    return true;
}

bool vr_app::setup_companion_window_pipeline() {
    companion_window_pipeline_ = make_graphics_pipeline(app_->device);

    if (!companion_window_pipeline_->add_shader(file_data("dpr/default_companion_vertex.spirv"), VK_SHADER_STAGE_VERTEX_BIT)) {
        return false;
    }

    if (!companion_window_pipeline_->add_shader(file_data("dpr/default_companion_fragment.spirv"), VK_SHADER_STAGE_FRAGMENT_BIT)) {
        return false;
    }

    companion_window_pipeline_->add_color_blend_attachment();

    companion_window_pipeline_->set_vertex_input_binding({ 0, sizeof(vertex), VK_VERTEX_INPUT_RATE_VERTEX });
    companion_window_pipeline_->set_vertex_input_attributes({
        { 0, 0, VK_FORMAT_R32G32B32_SFLOAT, to_ui32(offsetof(vertex, position)) },
        { 1, 0, VK_FORMAT_R32G32_SFLOAT, to_ui32(offsetof(vertex, uv)) },
    });

    companion_window_pipeline_descriptor_ = make_descriptor();
    companion_window_pipeline_descriptor_->add_binding(0, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT);
    companion_window_pipeline_descriptor_->add_binding(1, VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, VK_SHADER_STAGE_FRAGMENT_BIT);
    if (!companion_window_pipeline_descriptor_->create(app_->device)) {
        return false;
    }

    companion_window_pipeline_layout_ = make_pipeline_layout();
    companion_window_pipeline_layout_->add(companion_window_pipeline_descriptor_);
    if (!companion_window_pipeline_layout_ ->create(app_->device)) {
        return false;
    }

    companion_window_pipeline_descriptor_set_ = companion_window_pipeline_descriptor_->allocate(descriptor_pool_->get());

    VkSamplerCreateInfo const sampler_info = {
        .sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO,
        .magFilter = VK_FILTER_LINEAR,
        .minFilter = VK_FILTER_LINEAR,
        .mipmapMode = VK_SAMPLER_MIPMAP_MODE_LINEAR
    };
    if (!device()->vkCreateSampler(&sampler_info, &sampler)) {
        return false;
    }
    
    VkDescriptorImageInfo left_eye_image_descriptor_info = {
        .sampler = sampler,
        .imageView = framebuffers_[0]->color_image->get_view(),
        .imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
    };
    VkDescriptorImageInfo right_eye_image_descriptor_info = {
        .sampler = sampler,
        .imageView = framebuffers_[1]->color_image->get_view(),
        .imageLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL,
    };

    VkWriteDescriptorSet const write_desc_left_eye {
        .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
        .dstSet = companion_window_pipeline_descriptor_set_,
        .dstBinding = 0,
        .descriptorCount = 1,
        .descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
        .pImageInfo = &left_eye_image_descriptor_info,
    };

    VkWriteDescriptorSet const write_desc_right_eye {
        .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
        .dstSet = companion_window_pipeline_descriptor_set_,
        .dstBinding = 1,
        .descriptorCount = 1,
        .descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
        .pImageInfo = &right_eye_image_descriptor_info,
    };

    app_->device->vkUpdateDescriptorSets({ write_desc_left_eye, write_desc_right_eye });

    companion_window_pipeline_->on_process = [this](VkCommandBuffer command_buffer) {
        companion_window_pipeline_layout_->bind(command_buffer, companion_window_pipeline_descriptor_set_);
        quad_mesh_->bind_draw(command_buffer);
    };

    companion_window_pipeline_->set_layout(companion_window_pipeline_layout_);
    companion_window_pipeline_->create(app_->shading.get_pass()->get());
    app_->shading.get_pass()->add_front(companion_window_pipeline_);

    app_->add_run_end([this]() {
        companion_window_pipeline_descriptor_->free(companion_window_pipeline_descriptor_set_, descriptor_pool_->get());
        descriptor_pool_->destroy();
        companion_window_pipeline_descriptor_->destroy();
        companion_window_pipeline_->destroy();
        companion_window_pipeline_layout_->destroy();
        for (auto& strategy : stereo_strategies_) {
          strategy.reset();
        }
        device()->vkDestroySampler(sampler);
    });

    return true;
}


stereo_framebuffer::ptr vr_app::create_framebuffer(glm::uvec2 size, vr::EVREye eye) {
    stereo_framebuffer::ptr framebuffer = std::make_shared<stereo_framebuffer>();

    framebuffer->color_image = make_image(VK_FORMAT_R8G8B8A8_UNORM);
    framebuffer->color_image->set_usage(VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT);
    framebuffer->color_image->set_tiling(VK_IMAGE_TILING_OPTIMAL);
    if (!framebuffer->color_image->create(app_->device, size)) {
        log()->error("Failed to create color attachment");
        return {};
    }

    framebuffer->color_attachment = make_attachment(VK_FORMAT_R8G8B8A8_UNORM);
    framebuffer->color_attachment->set_op(VK_ATTACHMENT_LOAD_OP_CLEAR, VK_ATTACHMENT_STORE_OP_STORE);
    framebuffer->color_attachment->set_stencil_op(VK_ATTACHMENT_LOAD_OP_DONT_CARE, VK_ATTACHMENT_STORE_OP_DONT_CARE);
    framebuffer->color_attachment->set_layouts(VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL);

    framebuffer->depth_stencil_image = make_image(VK_FORMAT_D32_SFLOAT);
    framebuffer->depth_stencil_image->set_usage(VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT | VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT);
    framebuffer->depth_stencil_image->set_tiling(VK_IMAGE_TILING_OPTIMAL);
    if (!framebuffer->depth_stencil_image->create(app_->device, size)) {
        log()->error("Failed to create depth stencil attachment");
        return {};
    }

    framebuffer->depth_stencil_attachment = make_attachment(VK_FORMAT_D32_SFLOAT);
    framebuffer->depth_stencil_attachment->set_op(VK_ATTACHMENT_LOAD_OP_CLEAR, VK_ATTACHMENT_STORE_OP_STORE);
    framebuffer->depth_stencil_attachment->set_stencil_op(VK_ATTACHMENT_LOAD_OP_DONT_CARE, VK_ATTACHMENT_STORE_OP_DONT_CARE);
    framebuffer->depth_stencil_attachment->set_layouts(VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL);

    framebuffer->per_frame_data.view = glm::identity<glm::mat4>();

    framebuffer->per_frame_ubo = make_buffer();
    if (!framebuffer->per_frame_ubo->create_mapped(device(), &framebuffer->per_frame_data, sizeof(framebuffer->per_frame_data), VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)) {
        return {};
    }

    framebuffer->per_frame_descriptor_set = per_frame_descriptor_->allocate_set(descriptor_pool_->get());

    const VkWriteDescriptorSet per_frame_ubo_desc {
        .sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET,
        .dstSet = framebuffer->per_frame_descriptor_set,
        .dstBinding = 0,
        .descriptorCount = 1,
        .descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
        .pBufferInfo = framebuffer->per_frame_ubo->get_descriptor_info(),
    };
    device()->vkUpdateDescriptorSets({ per_frame_ubo_desc });
    return framebuffer;
}

void vr_app::calculate_projection_matrices() {
  calculate_projection_matrix(vr::EVREye::Eye_Left);
  calculate_projection_matrix(vr::EVREye::Eye_Right);
}

void vr_app::calculate_projection_matrix(vr::EVREye eye) {
    auto framebuffer = get_framebuffer_for_eye(eye);
    const float near_plane = app_->camera.z_near;
    const float far_plane = app_->camera.z_far;

    if (vr_system) {
        const vr::HmdMatrix44_t projection_matrix = vr_system->GetProjectionMatrix(eye, near_plane, far_plane);
        std::memcpy(&framebuffer->per_frame_data.projection[0][0], projection_matrix.m, sizeof(float) * 16);

        const vr::HmdMatrix34_t eye_to_head = vr_system->GetEyeToHeadTransform(eye);
        for (int i = 0; i < 4; ++i) {
            for (int j = 0; j < 3; ++j) {
                framebuffer->per_frame_data.headToEye[i][j] = eye_to_head.m[j][i];
            }
            framebuffer->per_frame_data.headToEye[i][3] = i == 3 ? 1.0f : 0.0f;
        }
        framebuffer->per_frame_data.projection = glm::transpose(framebuffer->per_frame_data.projection);
    } else {
        framebuffer->per_frame_data.projection = glm::perspective(glm::pi<float>() * 0.5f, 16.0f/9.0f, near_plane, far_plane);
        framebuffer->per_frame_data.headToEye = glm::translate(glm::mat4(1.0f), glm::vec3(10.0f, 0.0, 0.0f) * (eye == vr::Eye_Left ? 1.0f : -1.0f));
    }
}
